#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "xwin.h"
#include "ftread.h"
#include <X11/keysym.h>

static int nbands, linegap;

typedef struct {
  int nbands;
  int *y_adj;
  int *file_offsets;
  XImage **bands;
} Paragraph;

typedef struct {
  int which_paragraph;
  int which_band;
} Location;

static FILE *text_file;
static FILE *idx_file;
static int text_file_size;
static Paragraph *book;
int book_size;
static Location top_band;

static char *word=0;

static int
read_word(int idx, int idx_stop)
{
  static int wordmax=0;
  int i=0, dc=0;
  if (!word)
    {
      wordmax = 10;
      word = (char *)malloc(wordmax+1);
    }
  while (1)
    {
      int c = fgetc(text_file);
      dc ++;
      idx++;
      if (c == EOF || idx == idx_stop)
	{
	  word[i] = 0;
	  return dc;
	}
      if (!isgraph(c))
	{
	  if (i == 0)
	    continue;
	  word[i] = 0;
	  return dc;
	}
      if (i <= wordmax)
	{
	  wordmax += 10;
	  word = (char *)realloc(word, wordmax+1);
	}
      word[i++] = c;
    }
}

static int
read_para(int n, int where, int which)
{
  int rv = 0;
  int idx, idx_stop=-1;
  Paragraph *para = book + n;

  fseek (idx_file, n * sizeof(int), SEEK_SET);
  fread (&idx, 1, sizeof(int), idx_file);
  fread (&idx_stop, 1, sizeof(int), idx_file);
  fseek (text_file, idx, SEEK_SET);

  int max = 10;
  para->nbands = 0;
  para->bands = (XImage **)malloc(max * sizeof(XImage *));
  para->y_adj = (int *)malloc(max * sizeof(int));
  para->file_offsets = (int *)malloc((max+1) * sizeof(int));

  static int linemax = 100;
  static char *line = (char *)malloc(linemax+1);
  int nline = 0;
  int dc;
  int indent_spaces = string_width("m") / string_width(" ");
  while (1)
    {
      para->file_offsets[para->nbands] = idx;
      idx += read_word(idx, idx_stop);
      if (!word[0])
	break;
      int wordlen = strlen(word);
      if (nline + wordlen + 2 >= linemax)
	{
	  linemax = nline + wordlen + 2 + 100;
	  line = (char *)realloc(line, linemax+1);
	}
      if (nline == 0)
	{
	  for (nline=0; nline<indent_spaces; nline++)
	    line[nline] = ' ';
	  strcpy (line+nline, word);
	  nline += wordlen;
	  continue;
	}
      char *eol = line + nline;
      eol[0] = ' ';
      strcpy(eol+1, word);
      nline += wordlen+1;
      int sw = string_width (line);
      if (sw > win_width - 2*margin)
	{
	  /* Nope, won't fit */
	  eol[0] = 0;
	  if (para->nbands >= max-1)
	    {
	      max += 10;
	      para->bands = (XImage **)realloc(para->bands, max * sizeof(XImage *));
	      para->y_adj = (int *)realloc(para->y_adj, max * sizeof(int));
	      para->file_offsets = (int *)realloc(para->file_offsets, (max+1) * sizeof(int));
	    }
	  para->bands[para->nbands] = build_string(line, &(para->y_adj[para->nbands]));
	  para->nbands ++;
	  strcpy (line, word);
	  nline = wordlen;

	  if (font_is_freetype
	      && where != -1
	      && para->nbands-1 == which
	      && where < nbands)
	    {
	      XPutImage(display, window, gc, para->bands[which],
			0, 0, margin, where*linegap + para->y_adj[which],
			para->bands[which]->width, para->bands[which]->height);
	      XFlush(display);
	      which ++;
	      where ++;
	      rv ++;
	    }
	}
    }
  para->bands[para->nbands] = build_string(line, &(para->y_adj[para->nbands]));
  para->file_offsets[para->nbands+1] = idx;
  para->nbands ++;
  return rv;
}

static void
DHT(int where, char *string)
{
  int w = win_width/2 - XTextWidth(xfs, string, strlen(string))/2;
  XDrawString(display, window, gc, w, win_height*(where*2+1)/8, string, strlen(string));
}

static void
drawhelp()
{
  int i;
  int fh = xfs->ascent + xfs->descent;
  XClearArea(display, window, 0, 0, 0, 0, 0);

  XSetForeground(display, gc, grays[3]);
  XFillRectangle(display, window, gc, 0, 0, win_width/6, win_height/4);
  XFillRectangle(display, window, gc, win_width*5/6, 0, win_width/6, win_height);

  XSetForeground(display, gc, grays[5]);
  XDrawString(display, window, gc, 2, win_height/2-xfs->descent, TITLE, strlen(TITLE));
  XDrawString(display, window, gc, 2, win_height/2+xfs->ascent, COPYRIGHT, strlen(COPYRIGHT));

  XSetForeground(display, gc, grays[15]);
  XDrawString(display, window, gc, 2, win_height/8, "Done", 4);
  DHT(0, "PageUp");
  DHT(1, "LineUp");
  DHT(2, "LineDown");
  DHT(3, "PageDown");
  for (i=0; i<6; i++)
    {
      static char tmp[2] = {0, 0};
      tmp[0] = "SCROLL"[i];
      int w = XTextWidth(xfs, tmp, 1);
      XDrawString(display, window, gc, win_width*11/12-w/2, win_height/2 - (6-i*2)*fh/2, tmp, 1);
    }
  XDrawString(display, window, gc, 2, win_height-2-xfs->descent, "(Click to continue)", 19);
}

static int last_band_showing;

static void
redraw(int clear)
{
  if (!sawhelp)
    {
      drawhelp();
      return;
    }

  if (clear)
    {
      XClearArea(display, window, 0, 0, 0, 0, 0);
    }

  int x;
  int p, b;
  Paragraph *para;
  int first_offset=-1, last_offset=0;

  x=0;
  p = top_band.which_paragraph;
  para = book+p;
  b = top_band.which_band;

  if (!font_is_freetype)
    XSetFont(display, gc, text_fs->fid);
  XSetForeground(display, gc, grays[15]);

  last_band_showing = 0;

  while (x<nbands)
    {
      if (!para->bands)
	{
	  XFlush(display);
	  int done = read_para(p, x, b);
	  x += done;
	  b += done;
	  continue;
	}
      if (first_offset == -1)
	first_offset = para->file_offsets[b];
      last_offset = para->file_offsets[b+1];
      if (font_is_freetype)
	{
	  XPutImage(display, window, gc, para->bands[b],
		    0, 0, margin, x*linegap + para->y_adj[b], para->bands[b]->width, para->bands[b]->height);
	}
      else
	{
	  XDrawString(display, window, gc, margin, x*linegap + para->y_adj[b],
		      (char *)para->bands[b], strlen((char *)para->bands[b]));
	}
      x ++;
      b ++;
      if (b >= para->nbands)
	{
	  b = 0;
	  p++;
	  if (p == book_size)
	    {
	      last_band_showing = 1;
	      break;
	    }
	  para = book+p;
	}
    }
  int bar_top = win_height * first_offset / text_file_size;
  if (top_band.which_paragraph == 0
      && top_band.which_band == 0)
    bar_top = 0;
  int bar_bot = win_height * last_offset / text_file_size;
  XDrawLine(display, window, gc, win_width-1, bar_top, win_width-1, bar_bot);
}

static void
build_idx(char *tf, char *xf)
{
  FILE *textf, *idxfile;
  int idx=0, c, oc=2;

  textf = fopen(tf, "rb");
  idxfile = fopen(xf, "wb");

  while ((c = fgetc(textf)) != EOF)
    {
      if (oc && isgraph((unsigned char)c))
	{
	  if (oc > 1)
	    fwrite(&idx, 1, sizeof(int), idxfile);
	  oc = 0;
	}
      if (c == '\n')
	oc ++;
      idx++;
    }
  fclose(idxfile);
  fclose(textf);
}

void
do_scroll(int scroll)
{
  if (scroll > 0 && last_band_showing)
    return;
  while (scroll)
    {
      top_band.which_band += scroll;
      scroll = 0;
      if (top_band.which_band < 0)
	{
	  scroll = top_band.which_band;
	  top_band.which_paragraph --;
	  if (top_band.which_paragraph < 0)
	    {
	      top_band.which_paragraph = 0;
	      top_band.which_band = 0;
	      break;
	    }
	  top_band.which_band = book[top_band.which_paragraph].nbands;
	}
      else if (top_band.which_band >= book[top_band.which_paragraph].nbands)
	{
	  scroll = top_band.which_band - book[top_band.which_paragraph].nbands;
	  top_band.which_paragraph ++;
	  if (top_band.which_paragraph >= book_size)
	    {
	      top_band.which_paragraph = book_size-1;
	      top_band.which_band = book[top_band.which_paragraph].nbands-1;
	      break;
	    }
	  top_band.which_band = 0;
	}
    }
  redraw(1);
}

static int last_filepos;

static void
scroll_to_filepos(int filepos)
{
  int minp=-1, maxp=book_size, p, first, last, b;
  if (filepos == last_filepos)
    return;
  last_filepos = filepos;
  while (maxp > minp+1)
    {
      p = (minp+maxp)/2;
      fseek (idx_file, p * sizeof(int), SEEK_SET);
      fread (&first, 1, sizeof(int), idx_file);
      fread (&last, 1, sizeof(int), idx_file);
      if ((first <= filepos || p == 0)
	  && (last > filepos || p == book_size-1))
	{
	  if (!book[p].bands)
	    read_para(p, -1, -1);
	  for (b=0; b<book[p].nbands; b++)
	    if (book[p].file_offsets[b] >= filepos)
	      {
		top_band.which_paragraph = p;
		top_band.which_band = b ? b-1 : b;
		redraw(1);
		return;
	      }
	}
      if (first > filepos)
	maxp = p;
      else
	minp = p;
    }
  top_band.which_paragraph = book_size-1;
  if (!book[book_size-1].bands)
    read_para(book_size-1, -1, -1);
  top_band.which_band = book[book_size-1].nbands-1;
  redraw(1);
  return;
}

void
flow()
{
  int x, y;

  linegap = ascent + descent;
  nbands = win_height / linegap;

  idxfile = (char *)malloc(strlen(textfile) + strlen(IDXSUFF) + 1);
  strcpy(idxfile, textfile);
  strcat(idxfile, IDXSUFF);

  struct stat ts, s;
  stat (textfile, &ts);
  if (stat (idxfile, &s) < 0 || ts.st_mtime > s.st_mtime)
    build_idx (textfile, idxfile);
  if (stat (idxfile, &s) < 0)
    return;
  text_file_size = ts.st_size;
  book_size = s.st_size / sizeof(int);
  book = (Paragraph *)malloc (book_size * sizeof(Paragraph));
  memset (book, 0, book_size * sizeof(Paragraph));
  top_band.which_paragraph = 0;
  top_band.which_band = 0;

  text_file = fopen(textfile, "rb");
  idx_file = fopen(idxfile, "rb");

  load_font();

  redraw(1);

  int scrolling = 0;
  last_filepos = -1;
  XEvent e;
  while (1)
    {
      xwin_event(&e);
      switch (e.type)
	{
	case Expose:
	  redraw(0);
	  break;
	case ButtonPress:
	  if (!sawhelp)
	    {
	      sawhelp = 1;
	      redraw(1);
	      break;
	    }
	  x = e.xbutton.x * 6 / win_width;
	  y = e.xbutton.y * 4 / win_height;
	  if (x == 5)
	    {
	      scroll_to_filepos(text_file_size * e.xbutton.y / win_height);
	      scrolling = 1;
	      break;
	    }
	  scrolling = 0;
	  if (x == 0 && y == 0)
	    {
	      fclose(text_file);
	      fclose(idx_file);
	      return;
	    }
	  int scroll;
	  switch (y)
	    {
	    case 0: scroll = -nbands; break;
	    case 1: scroll = -1; break;
	    case 2: scroll = 1; break;
	    case 3: scroll = nbands; break;
	    }
	  do_scroll(scroll);
	  break;
	case MotionNotify:
	  if (scrolling)
	    {
	      if (e.xmotion.y < 0) e.xmotion.y = 0;
	      if (e.xmotion.y >= win_height) e.xmotion.y = win_height-1;
	      scroll_to_filepos(text_file_size * e.xmotion.y / win_height);
	    }
	  break;
	case KeyPress:
	  char c;
	  KeySym key;
	  if (XLookupString(&e.xkey, &c, 1, &key, 0) == 1)
	    key = c;
	  scroll = 0;
	  switch (key)
	    {
	    case XK_Page_Up:
	    case XK_Left:	scroll = -nbands; break;
	    case XK_Up:		scroll = -1; break;
	    case XK_Down:	scroll = 1; break;
	    case XK_Page_Down:
	    case XK_Right:	scroll = nbands; break;
	    case 'q':
	    case 'Q':
	    case 8:
	    case 127:
	    case XK_Delete:
	    case XK_KP_Delete:
	    case XK_BackSpace:	return;
	    }
	  if (scroll)
	    do_scroll(scroll);
	  break;
	}
    }
}
