#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "xwin.h"
#include "ftread.h"
#include "rc.h"
#include "fileselector.h"
#include <X11/keysym.h>

static int done;

static char *texttail, *fonttail;
static int text_y, font_y, size_y;
static XImage *sample;

static void
arrow (int where, int points, int count)
{
  int x, y, i, d;
  y = win_height * 19 / 32;
  x = win_width * (where+10) / 20;
  d = points * xfs->ascent;
  for (i=0; i<count; i++)
    {
      XDrawLine(display, window, gc, x, y, x+d, y+xfs->ascent);
      XDrawLine(display, window, gc, x, y, x+d, y-xfs->ascent);
      x -= d;
    }
}

static void
aline (int which, int x1, int y1, int x2, int y2)
{
  which *= win_width*3/4;
  x1 = win_width*x1/400 + which;
  y1 = win_height*y1/400 + win_height/4;
  x2 = win_width*x2/400 + which;
  y2 = win_height*y2/400 + win_height/4;
  XDrawLine(display, window, gc, x1, y1, x2, y2);
}

static void
redraw(int clear)
{
  int fh = xfs->ascent + xfs->descent;
  int fy = xfs->ascent - fh/2;

  if (clear)
    {
      if (fontsize < 4) fontsize = 4;
      if (fontsize > 99) fontsize = 99;
      XClearArea(display, window, 0, 0, 0, 0, 0);
      load_font();
      if (font_is_freetype);
      sample = build_string("Read...", 0);
    }

  XSetForeground(display, gc, grays[5]);

  XDrawString(display, window, gc, 2, 2+xfs->ascent, TITLE, strlen(TITLE));
  XDrawString(display, window, gc, 2, 2+2*xfs->ascent+xfs->descent, COPYRIGHT, strlen(COPYRIGHT));

#define VLINE(x) XDrawLine(display, window, gc, 0, x, win_width, x);
  VLINE (win_height*1/4);
  VLINE (win_height*2/4);
  VLINE (win_height*11/16);
  VLINE (win_height - xfs->ascent - xfs->descent - 4);
  if (font_is_freetype)
    XDrawLine(display, window, gc,
	      win_width*1/4, win_height*1/4,
	      win_width*1/4, win_height*2/4);
  else
    XDrawLine(display, window, gc,
	      win_width*3/4, win_height*1/4,
	      win_width*3/4, win_height*2/4);
  if (font_is_freetype || font_is_scalable)
    {
      arrow(-8,  1, 2);
      arrow(-4,  1, 1);
      arrow( 4, -1, 1);
      arrow( 8, -1, 2);
    }
  aline(0, 20, 20, 80, 80);
  aline(0, 20, 80, 80, 20);
  aline(1, 20, 20, 80, 20);
  aline(1, 50, 20, 50, 80);

  XSetForeground(display, gc, grays[3]);
  XFillRectangle(display, window, gc, 0, win_height - xfs->ascent - xfs->descent - 4,
		 win_width, xfs->ascent + xfs->descent + 4);

  XSetForeground(display, gc, grays[15]);

  int c = win_width / 2;

  static int sample_y = 0;
  if (sample_y == 0)
    sample_y = (win_height*11/16
		+ win_height - xfs->ascent - xfs->descent - 4) / 2;

  if (font_is_freetype)
    XPutImage (display, window, gc, sample, 0, 0,
	       c - sample->width/2, sample_y - sample->height/2,
	       sample->width, sample->height);
  else
    {
      XSetFont(display, gc, text_fs->fid);
      XDrawImageString(display, window, gc,
		       c-string_width("Read...")/2,
		       ((win_height*11/16 + win_height - xfs->ascent - xfs->descent - 2)/2
			- (text_fs->ascent+text_fs->descent)/2
			+ ascent),
		       "Read...", 7);
    }

  XSetFont(display, gc, xfs->fid);
  int w = XTextWidth(xfs, texttail, strlen(texttail));
  XDrawImageString(display, window, gc, c-w/2, text_y+fy, texttail, strlen(texttail));

  w = XTextWidth(xfs, fonttail, strlen(fonttail));
  XDrawImageString(display, window, gc, c-w/2, font_y+fy, fonttail, strlen(fonttail));

  if (font_is_freetype || font_is_scalable)
    {
      char tmp[20];
      sprintf(tmp, " %d ", fontsize);
      w = XTextWidth(xfs, tmp, strlen(tmp));
      XDrawImageString(display, window, gc, c-w/2, size_y+fy, tmp, strlen(tmp));
    }

  w = XTextWidth(xfs, "QUIT", 4);
  XDrawString(display, window, gc, c-w/2, win_height-xfs->descent-1, "QUIT", 4);
}

void
figure_tails()
{
  char *c;

  texttail = textfile;
  for (c=textfile; *c; c++)
    if (*c == '/')
      texttail = c+1;
  fonttail = fontfile;
  if (*fonttail == '/')
    for (c=fontfile; *c; c++)
      if (*c == '/')
	fonttail = c+1;
}

void
select_screen()
{

  text_y = win_height * 1/8;
  font_y = win_height * 3/8;
  size_y = win_height * 19/32;

  figure_tails();

  redraw(1);

  XEvent e;
  while (!done)
    {
      xwin_event(&e);
      switch (e.type)
	{
	case Expose:
	  redraw(0);
	  break;
	case ButtonPress:
	  if (e.xbutton.y < win_height*1/4)
	    {
	      free (textfile);
	      textfile = strdup(file_selector(".txt"));
	      figure_tails();
	      redraw(1);
	    }
	  else if (e.xbutton.y < win_height*2/4)
	    {
	      free (fontfile);
	      if (e.xbutton.x < win_width/5)
		font_is_freetype = 0;
	      else if (e.xbutton.x > win_width*4/5)
		font_is_freetype = 1;
	      if (font_is_freetype)
		fontfile = strdup(file_selector(".ttf"));
	      else
		fontfile = strdup(font_selector());
	      figure_tails();
	      redraw(1);
	    }
	  else if (e.xbutton.y < win_height*11/16)
	    {
	      int y = e.xbutton.x / (win_width/4);
	      switch (y)
		{
		case 0:
		  fontsize -= 10;
		  break;
		case 1:
		  fontsize --;
		  break;
		case 2:
		  fontsize ++;
		  break;
		case 3:
		  fontsize += 10;
		  break;
		}
	      redraw(1);
	    }
	  else if (e.xbutton.y > win_height - xfs->ascent - xfs->descent - 4)
	    {
	      return;
	    }
	  else
	    {
	      write_rc_file(1);
	      flow();
	      redraw(1);
	    }
	  break;
	case KeyPress:
	  char c;
	  KeySym key;
	  if (XLookupString(&e.xkey, &c, 1, &key, 0) == 1)
	    key = c;
	  switch (key)
	    {
	    case XK_Left:
	      return;
	    case XK_Right:
	      write_rc_file(1);
	      flow();
	      redraw(1);
	      break;
	    case 'q':
	    case 'Q':
	      write_rc_file(1);
	      return;
	    }
	  break;
	}
    }
}
