#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "freetype/freetype.h"

FT_Library ftlib;
FT_Face face;

#define FLOOR(x) ((x) & ~63)

static int gammas[256];

main(int argc, char **argv)
{
  char *fontname = argc > 1 ? argv[1] : "comic.ttf";
  int pixsize = argc > 2 ? atoi(argv[2]) : 16;
  char *string = argc > 3 ? argv[3] : "Hello";
  fprintf(stderr, "usage: test1 <font.ttf> <ppem> <string> > foo.pgm\n");

  int error;

  for (int g = 0; g<256; g++)
    {
      double d = g / 255;
      d = pow (1.0 - d, 2);
      gammas[g] = (int)(d * 255);
      
    }
  
  if (FT_Init_FreeType (&ftlib))
    {
      fprintf(stderr, "can't initialize ftlib\n");
      exit(1);
    }

  error = FT_New_Face (ftlib, fontname, 0, &face);
  if (error == FT_Err_Unknown_File_Format)
    {
      fprintf(stderr, "unknown file format %s\n", fontname);
      exit(1);
    }
  else if (error)
    {
      fprintf(stderr, "can't open file %s\n", fontname);
      exit(1);
    }
  
  error = FT_Set_Pixel_Sizes (face, 0, pixsize);

#if 0
  if (FT_HAS_KERNING (face))
    {
      int i, j;
      FT_Vector kerning;
      fprintf(stderr, "has kerning\n");
      for (i=0; i<256; i++)
	{
	  int ii =  FT_Get_Char_Index(face, i);
	  for (j=0; j<256; j++)
	    {
	      kerning.x = kerning.y = 0;
	      FT_Get_Kerning (face, ii, FT_Get_Char_Index(face, j), 0, &kerning);
	      if (kerning.x || kerning.y)
		fprintf(stderr, "kern %c%c x %d y %d\n", i, j, kerning.x, kerning.y);
	    }
	}
    }
#endif

  int minx=0, miny=0, maxx=0, maxy=0;
  int penx=0, peny=0;
  int i;
  FT_GlyphSlot slot = face->glyph;

  for (i=0; string[i]; i++)
    {
      FT_Load_Char (face, string[i], FT_LOAD_RENDER);

      int px = (penx + 32) & ~63;
      FT_Glyph_Metrics *m = &(slot->metrics);
      
      FT_Vector kerning;
      if (i > 0 && 0 == FT_Get_Kerning (face,
					FT_Get_Char_Index(face, string[i-1]),
					FT_Get_Char_Index(face, string[i]), 0, &kerning))
	{
	  if (kerning.x || kerning.y)
	    fprintf(stderr, "kern %c%c x %d y %d\n", string[i-1], string[i], kerning.x, kerning.y);
	  penx += kerning.x;
	  peny += kerning.y;
	}

      if (minx > penx + m->horiBearingX)
	minx = penx + m->horiBearingX;
      if (maxx < penx + m->horiBearingX + m->width)
	maxx = penx + m->horiBearingX + m->width;

      if (miny > m->horiBearingY - m->height)
	miny = m->horiBearingY - m->height;
      if (maxy < m->horiBearingY)
	maxy = m->horiBearingY;

      penx += slot->advance.x;
      peny += slot->advance.y;
    }

  fprintf(stderr, "minx %d maxx %d miny %d maxy %d\n",
	  minx/64, maxx/64, miny/64, maxy/64);
  fprintf(stderr, "ascent %d descent %d height %d\n",
	  face->size->metrics.ascender/64,
	  face->size->metrics.descender/64,
	  face->size->metrics.height/64);
  if (miny > face->size->metrics.descender)
    miny = face->size->metrics.descender;
  if (maxy < face->size->metrics.ascender)
    maxy = face->size->metrics.ascender;
  if (maxx < penx)
    maxx = penx;

  fprintf(stderr, "global em %d bb %d - %d\n",
	  face->units_per_EM,
	  face->bbox.xMin * pixsize / face->units_per_EM,
	  face->bbox.xMax * pixsize / face->units_per_EM);

  int w = (maxx - minx)/64 + 2;
  int h = (maxy - miny)/64 + 2;

  unsigned char *pgm = (unsigned char *)malloc (w * h);
  penx = -minx;
  peny = -miny;
  memset (pgm, 255, w*h);

  for (i=0; string[i]; i++)
    {
      FT_Load_Char (face, string[i], FT_LOAD_RENDER);

      FT_Vector kerning;
      if (i > 0 && 0 == FT_Get_Kerning (face,
					FT_Get_Char_Index(face, string[i-1]),
					FT_Get_Char_Index(face, string[i]), 0, &kerning))
	{
	  if (kerning.x || kerning.y)
	    fprintf(stderr, "kern %c%c x %d y %d\n", string[i-1], string[i], kerning.x, kerning.y);
	  penx += kerning.x;
	  peny += kerning.y;
	}

      int left = penx/64 + slot->bitmap_left;
      int top = peny/64 + slot->bitmap_top;
      int dy, dr;
      unsigned char *row = slot->bitmap.buffer;
      int x, y, nrows;

      if (slot->bitmap.pitch > 0)
	{
	  dy = 1;
	  dr = slot->bitmap.pitch;
	  y = top;
	}
      else
	{
	  dy = -1;
	  dr = -slot->bitmap.pitch;
	  y = top - slot->bitmap.rows + 1;
	}
      y = h - y - 1;

      for (nrows = 0; nrows < slot->bitmap.rows; nrows++)
	{
	  for (x=0; x<slot->bitmap.width; x++)
	    pgm[left+x + w*y] -= ((unsigned char *)row)[x];

	  y += dy;
	  row += dr;
	}
      penx += slot->advance.x;
      peny += slot->advance.y;
    }
  printf("P5\n%d %d 255\n", w, h);
  fwrite(pgm, 1, w*h, stdout);
  return 0;
}
